/*

GS enctype2 servers list decoder 0.1
by Luigi Auriemma
e-mail: aluigi@autistici.org
web:    http://aluigi.altervista.org


INTRODUCTION
============

This is the algorithm used to decode the data sent by the Gamespy
master server when the client (you) requests the enctype 2 method.
Who wanna know the servers list does NOT need to use this enctype 2
method but the "classical" in which the servers list is not encoded.

Probably this algorithm is useless and needed by almost nobody but
it is here, is opensource and works perfectly.


USAGE
=====

    u_char  total_packet[] = "blablabla...bla",
            *data;
    int     size;

    size = sizeof(total_packet) - 1;

    data = enctype2_decoder(
        gamekey,
        total_packet,
        &size);

Variables:
----------
total_packet the total data received by the master server after the
             sending of our request, it contains the servers list
data         a pointer that will point where the servers list starts
size         it must contain the size of the packet and after the
             function it will contain the size of the server list
gamekey      the gamekey of the game, for example that of RogerWilco
             is "rW17Ko".

enctype2_decoder() naturally is my magic function that decodes the
servers list.



LICENSE
=======
    Copyright 2004 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt

*/



u_char *enctype2_decoder(u_char *key, u_char *data, int *size);
void enc2_40C410(u_char *src, int size, u_int *dest);
void enc2_40C160(u_int *data, int num1, int num2);
void enc2_40C4D0(u_int *A8, u_char *Ac, int A10);
void enc2_40C2A0(u_int *A4, u_int *A8, u_int Ac);



u_char *enctype2_decoder(u_char *key, u_char *data, int *size) {
    u_char  *dataptr;
    int     len,
            i;
    u_int   dest[325];

    *data ^= 0xec;

    len = strlen(key);
    dataptr = data + 1;
    for(i = 0; i < len; i++) dataptr[i] ^= key[i];

    for(i = 0; i < 325; i++) dest[i] = 0x00000000;

    enc2_40C410(dataptr, *data, dest);
    dataptr += *data;
    *size -= (*data + 1);
    enc2_40C4D0(dest, dataptr, *size);

    *size -= 6;
    return(dataptr);
}



void enc2_40C410(u_char *src, int size, u_int *dest) {
    int     i,
            x,
            y;
    u_char  cl;
    u_int   old;

    //  for(i = 0; i < 256; i++) dest[i] = 0;

    for(y = 0; y < 4; y++) {
        for(i = 0; i < 256; i++) {
            dest[i] = (dest[i] << 8) + i;
        }

        for(cl = y, x = 0; x < 2; x++) {
            for(i = 0; i < 256; i++) {
                old = dest[i];
                cl += old + src[i % size];
                dest[i] = dest[cl];
                dest[cl] = old;
            }
        }
    }

    for(i = 0; i < 256; i++) dest[i] ^= i;

    enc2_40C160(dest, 0, 0);
}



void enc2_40C160(u_int *data, int num1, int num2) {
    u_int   eax,
            ebx,
            edx,
            esi,
            edi;

    edx = num1;
    eax = 0;
    ebx = 0x8000;
    edi = 1;
    data[0x130] = 0;
    do {
        edx += edi;
        eax += edx;
        edx += eax;
        if(num2 & ebx) {
            esi = edx = ~edx;
            edi = (edi << 1) + 1;
            esi <<= 0x18;
            esi |= edx >> 8;
            esi ^= data[esi & 0xff];
            eax ^= data[eax & 0xff];
            edx = esi << 0x18;
            edx |= esi >> 8;
            esi = eax >> 0x18;
            esi |= eax << 8;
            edx ^= data[edx & 0xff];
            esi ^= data[esi & 0xff];
            eax = esi >> 0x18;
            eax |= esi << 8;
        } else {
            data[data[0x130] + 0x100] = edx;
            data[data[0x130] + 0x110] = eax;
            data[data[0x130] + 0x120] = edi;
            data[0x130]++;
            esi = eax << 0x18;
            esi |= eax >> 8;
            edx ^= data[edx & 0xff];
            esi ^= data[esi & 0xff];
            eax = esi << 0x18;
            eax |= esi >> 8;
            esi = edx >> 0x18;
            esi |= edx << 8;
            esi ^= data[esi & 0xff];
            eax ^= data[eax & 0xff];
            edx = esi >> 0x18;
            edx |= esi << 8;
            edi <<= 1;
        }
        ebx >>= 1;
    } while(ebx > 0);
    data[0x133] = edi;
    data[0x134] = num1;
    data[0x132] = eax;
    data[0x131] = edx;
    eax ^= edx;
}



void enc2_40C4D0(u_int *A8, u_char *Ac, int A10) {
    int     i;
    u_char  *ptr,
            *limit;

    ptr = limit = (u_char *)(A8 + 0x135);
    enc2_40C2A0(A8, (u_int *)ptr, 16);

    for(i = 0; i < A10; i++) {
        if((ptr - limit) >= 0x3f) {
            ptr = limit;
            enc2_40C2A0(A8, (u_int *)ptr, 16);
        }
        Ac[i] ^= *ptr;
        ptr++;
    }
}



void enc2_40C2A0(u_int *A4, u_int *A8, u_int Ac) {
    u_int   *ebx,
            edx,
            ebx1,
            eax,
            ecx,
            edi,
            ebp,
            *tmp;

    ebp = A4[0x130];
    eax = A4[0x131];
    ecx = A4[0x132];
    edi = A4[0x133];
    tmp = A8 + Ac;
    while(A8 < tmp) {
        if(edi < 0x10000) {
            ebx = A4 + ebp + 0x110;
            do {
                eax += edi;
                ebx++;
                ecx += eax;
                eax += ecx;
                ebx[-17] = eax;
                ebx[-1] = ecx;
                edx = ecx << 0x18;
                edx |= ecx >> 8;
                ecx = eax & 0xff;
                ebx[15] = edi;
                edi <<= 1;
                ebp++;
                eax ^= A4[ecx];
                edx ^= A4[edx & 0xff];
                ecx = edx << 0x18;
                ecx |= edx >> 8;
                edx = eax >> 0x18;
                edx |= eax << 8;
                edx ^= A4[edx & 0xff];
                ecx ^= A4[ecx & 0xff];
                eax = edx >> 0x18;
                eax |= edx << 8;
            } while(edi < 0x10000);
        }
        ecx ^= eax;
        *A8++ = ecx;
        ebp--;
        eax = A4[ebp + 0x100];
        edi = A4[ebp + 0x110];
        eax = ~eax;
        ecx = eax << 0x18;
        ecx |= eax >> 8;
        eax = edi;
        edx = ecx;
        eax &= 0xff;
        ecx ^= A4[edx & 0xff];
        ebx1 = A4[eax];
        eax = ecx;
        edi ^= ebx1;
        eax <<= 0x18;
        eax |= ecx >> 8;
        edx = edi;
        ecx = eax;
        edx >>= 0x18;
        edi <<= 8;
        ecx &= 0xff;
        edx |= edi;
        ebx1 = A4[ecx];
        ecx = edx & 0xff;
        eax ^= ebx1;
        edx ^= A4[ecx];
        ecx = edx >> 0x18;
        ecx |= edx << 8;
        edi = (A4[ebp + 0x120] << 1) + 1;
    }
    A4[0x133] = edi;
    A4[0x130] = ebp;
    A4[0x131] = eax;
    A4[0x132] = ecx;
}



