/*

======
Gslist
======

INTRODUCTION
------------
Gslist is a game servers browser and heartbeats sender, it supports an
incredible amount of games and moreover a lot of options making it
simple and complete at the same time.
Then it is opensource and can be compiled on any system with small
changes (if needed).

All this software is based on the FREE and OFFICIAL PUBBLIC data
available online just by the same Gamespy on their PUBBLIC servers.


CONTRIBUTORS
------------
- Steve Hartland:  filters (cool) and FREEBSD compatibility (netinet/in.h)
- Cimuca Cristian: bug reports and ideas
- Ludwig Nussel:   gslist user's directory (very useful in Unix), quiet
                   option and other tips


LICENSE
-------
    Copyright 2004,2005 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close       closesocket
    #define ONESEC      1000
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #ifndef strcasestr  // avoid warnings
        extern char *strcasestr (__const char *__haystack, __const char *__needle)
     __THROW __attribute_pure__;
    #endif

    #define ONESEC      1
    #define stristr     strcasestr
    #define strnicmp    strncasecmp
#endif



#define VER         "0.6.3"
#define MS          "master.gamespy.com"
#define MSPORT      28900
#define HBPORT      27900
#define BUFFSZ      2048
#define TOTSZ       256
#define GSHB1       "\\heartbeat\\%hu" \
                    "\\gamename\\%s"
#define GSHB2a      "\\validate\\%s" \
                    "\\final\\" \
                    "\\queryid\\1.1"
#define GSHB2b      "\\gamename\\%s" \
                    "\\final\\" \
                    "\\queryid\\1.1"
#define PCK         "\\gamename\\%s" \
                    "\\enctype\\%d" \
                    "\\validate\\%s" \
                    "\\final\\" \
                    "\\queryid\\1.1" \
                    "\\list\\cmp" \
                    "\\gamename\\%s" \
                    "%s%s" \
                    "\\final\\"
#define HOST        "motd.gamespy.com"
#define PORT        80
#define H416        "416"
#define H200        "200"
#define H206        "206"
#define GETSZ       8192
#define NAMESZ      64
#define ENDHEAD     "\r\n\r\n"
#define CLEN        "Content-Length: "
#define GETNAME     "GET /software/services/index.aspx?mode=modified HTTP/1.0\r\n" \
                    "Host: "HOST"\r\n" \
                    "\r\n"
#define GET1        "GET /software/services/index.aspx?mode=full&services="
#define GET2        " HTTP/1.0\r\n" \
                    "Host: "HOST"\r\n" \
                    "\r\n"
#define TIMEOUT     3
#define HBMINUTES   300  // 300 seconds = 5 minutes
#define FILEOUT     "gslist-out.gsl"
#define CFG         "gslist.cfg"
#define FULLCFG     "full.cfg"
#define KNOWNCFG    "knownsvc.cfg"
#define OLDGSINFO   "\\status\\"
#define MAXNAMEREQ  80
#define MULTIQUERY  \
        "-d T HOST PORT another option for various queries. The T parameter is referred\n" \
        "               to the type of query you want to use:\n" \
        "                 1 = Quake 3 engine    2 = Medal of Honor    3 = Quake 2 engine\n" \
        "                 4 = Half-life         5 = DirectPlay8       6 = Doom3\n" \
        "                 7 = All-Seeing Eye\n" \
        "                 ? = shows the list of supported queries (no HOST/PORT needed)\n"

#define CHK_BUFFLEN if((len < 0) || (len > BUFFSZ)) { \
                        fprintf(stderr, \
                            "\n" \
                            "Error: the used buffer is smaller than how much needed (%d < %d)\n" \
                            "\n", len, BUFFSZ); \
                        exit(1); \
                    }
#define CNAMEOFF    7
#define CFNAMEOFF   23
#define CKEYOFF     73
#define CSPACES     50
#define CNAMELEN    (CFNAMEOFF - CNAMEOFF)
#define CFNAMELEN   (CKEYOFF - CFNAMEOFF)



void gs_info_udp(u_int ip, u_short port, int mode);
int check_updates(void);
void gsfullup(void);
void http2file(u_char *getstr, int len, FILE *fd);
void write_err(void);
void build_cfg(void);
void get_key(int gamenum, u_char *gamestr, u_char *gamekey);
void show_list(u_char *name);
void show_countries(void);
void show_help(void);
void show_examples(void);
void show_gslist_title(void);
int timeout(int sock);
u_int resolv(char *host);
void std_err(void);
FILE *gslfopen(const char *path, const char *mode);



FILE    *fdout;     // for the -o option
int     quiet = 0;



#include "enctype2_decoder.h"   // for enctype2 decoding
#include "gsmsalg-old2.h"       // the old 0.2.1 version since it's easier to use
#include "show_dump.h"          // hex dump of data (used in multi_query)
#include "multi_query.h"        // to use the -d option



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer,
                        peerl;
#pragma pack(2)
    struct  ipport_struct {
                u_int   ip;
                u_short port;
            } *ipport;
#pragma pack()
    u_int   servers,
            msip;
    int     sd,
            err,
            len,
            i,
            enctype          = 0,
            execlen          = 0,
            hbmethod         = 0,
            psz,
            dynsz,
            itsok            = 0,
            on               = 1;
    u_short msport           = MSPORT,
            heartbeat_port   = 0;
    u_char  *buff            = NULL,
            gamekey[7],
            gamestr[20],
            *sec             = NULL,
            *key             = NULL,
            *ms              = MS,
            outtype          = 0,
            *tmpexec         = NULL,
            *execstring      = NULL,
            *execstring_ip   = NULL,
            *execstring_port = NULL,
            *filter          = "",
            *where           = "",
            *execptr         = NULL;

#ifdef WIN32
    WSADATA wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    if(argc < 2) {
        show_help();
        exit(1);
    }

    fdout = stdout;
    *gamestr = 0;
    *gamekey = 0;

    for(i = 1; i < argc; i++) {
        if(argv[i][0] != '-') {
            fprintf(stderr, "\n"
                "Error: recheck your options (%s is not valid)\n"
                "\n", argv[i]);
            exit(1);
        }

        switch(argv[i][1]) {
            case '-':
            case '?':
            case 'h': {
                show_help();
                return(0);
                } break;
            case 'e': {
                show_examples();
                return(0);
                } break;
            case 'N': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you MUST select a GAMENAME!\n"
                        "Use -l or -s for the full list or searching a specific one\n"
                        "\n", stderr);
                    exit(1);
                }
                strncpy(gamestr, argv[i], sizeof(gamestr) - 1);
                get_key(0, gamestr, gamekey);
                } break;
            case 'n': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you MUST select a game number!\n"
                        "Use -l or -s for the full list or searching a specific one\n"
                        "\n", stderr);
                    exit(1);
                }
                get_key(atoi(argv[i]), gamestr, gamekey);
                } break;
            case 'l': {
                show_list(0);
                return(0);
                } break;
            case 's': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you MUST specify a text pattern to search in the game database\n"
                        "\n", stderr);
                    exit(1);
                }
                show_list(argv[i]);
                return(0);
                } break;
            case 'u': {
                gsfullup();
                return(0);
                } break;
            case 'i': {
                i++;
                if(!argv[i] || !argv[i + 1]) {
                    fputs("\n"
                        "Error: you MUST insert an host and a query port\n"
                        "\n", stderr);
                    exit(1);
                }
                gs_info_udp(resolv(argv[i]), atoi(argv[i + 1]), 0);
                return(0);
                } break;
            case 'I': {
                i++;
                if(!argv[i] || !argv[i + 1]) {
                    fputs("\n"
                        "Error: you MUST insert an host and a query port\n"
                        "\n", stderr);
                    exit(1);
                }
                gs_info_udp(resolv(argv[i]), atoi(argv[i + 1]), 1);
                return(0);
                } break;
            case 'd': {
                i++;
                if(!argv[i] || (argv[i][0] == '?')) {
                    fputs(MULTIQUERY, stdout);
                    return(0);
                }
                if(!argv[i] || !argv[i + 1] || !argv[i + 2]) {
                    fputs("\n"
                        "Error: you MUST insert the type of query, a host and a query port\n"
                        "\n", stderr);
                    exit(1);
                }
                multi_query(atoi(argv[i]), resolv(argv[i + 1]), atoi(argv[i + 2]));
                return(0);
                } break;
            case 'f': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you MUST specify a filter\n"
                        "\n", stderr);
                    exit(1);
                }
                filter = argv[i];
                where = "\\where\\";
                } break;
            case 'c': {
                show_countries();
                return(0);
                } break;
            case 'r': {
                i++;
                execstring = argv[i];
                execlen = strlen(execstring) + 23;

                tmpexec = malloc(execlen);
                if(!tmpexec) std_err();

                execstring_ip = strstr(execstring, "#IP");
                execstring_port = strstr(execstring, "#PORT");
                if(execstring_ip) *execstring_ip = 0x00;
                if(execstring_port) *execstring_port = 0x00;

                execlen = strlen(execstring);
                memcpy(tmpexec, execstring, execlen);
                } break;
            case 'b': {
                i++;
                msport = HBPORT;
                heartbeat_port = atoi(argv[i]);
                hbmethod = 1;
                } break;
            case 'B': {
                i++;
                msport = HBPORT;
                heartbeat_port = atoi(argv[i]);
                hbmethod = 2;
                } break;
            case 'x': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you have forgotten some parameters\n"
                        "\n", stderr);
                    exit(1);
                }
                ms = strchr(argv[i], ':');
                if(ms) {
                    msport = atoi(ms + 1);
                    *ms = 0x00;
                }
                ms = argv[i];
                } break;
            case 'o': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you have forgotten the type of output or the filename\n"
                        "\n", stderr);
                    exit(1);
                }
                outtype = atoi(argv[i]);
                if(outtype > 6) outtype = 0;
                if(!outtype) {
                    fdout = fopen(argv[i], "wb");
                    if(!fdout) std_err();
                }
                } break;
            case 'q': {
                quiet = 1;
                } break;
            case 'y': {
                i++;
                if(!argv[i] || !argv[i + 1]) {
                    fputs("\n"
                        "Error: you MUST chose a gamename and a game key!\n"
                        "\n", stderr);
                    exit(1);
                }
                strncpy(gamestr, argv[i], sizeof(gamestr) - 1);
                gamestr[sizeof(gamestr) - 1] = 0x00;
                i++;
                strncpy(gamekey, argv[i], sizeof(gamekey) - 1);
                gamekey[sizeof(gamekey) - 1] = 0x00;
                } break;
            case 't': {
                i++;
                if(!argv[i]) {
                    fputs("\n"
                        "Error: you MUST select an enctype number!\n"
                        "\n", stderr);
                    exit(1);
                }
                enctype = atoi(argv[i]);
                } break;
            case 'v': {
                show_gslist_title();
                fputs("Project home: http://aluigi.altervista.org/papers.htm#gslist\n\n", stdout);
                return(0);
                } break;
            default: {
                fprintf(stderr, "\n"
                    "Error: wrong argument (%s)\n"
                    "\n", argv[i]);
                exit(1);
            } break;
        }
    }

    if(!*gamekey || !*gamestr) {
        fputs("\n"
            "Error: The game is not available or has not been specified.\n"
            "       A valid game must have both gamename and gamekey.\n"
            "       Probably you have only forgotten to use the -N, -n or -y option\n"
            "\n", stderr);
        exit(1);
    }

    if(!quiet) {
        fprintf(fdout,
            "\n"
            "Gamename:    %s    Key: %s\n"
            "Filter:      %s\n"
            "Enctype:     %d\n"
            "Resolving    %s ...\n",
            gamestr, gamekey,
            filter,
            enctype,
            ms);
    }

    peer.sin_addr.s_addr = msip = resolv(ms);
    peer.sin_port        = htons(msport);
    peer.sin_family      = AF_INET;

    buff = malloc(BUFFSZ + 1);
    if(!buff) std_err();
    dynsz = BUFFSZ;

    if(!quiet) {
        fprintf(fdout,
            "Server:      %s:%hu\n",
            inet_ntoa(peer.sin_addr),
            msport);
    }

    if(hbmethod == 2) {
        printf("\n"
            "The heartbeat for port %d will be sent each 5 minutes\n",
            heartbeat_port);

        len = snprintf(
            buff,
            BUFFSZ,
            GSHB1,
            heartbeat_port,
            gamestr);
        CHK_BUFFLEN;

        for(;;) {
            sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
            if(sd < 0) std_err();

            if(sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(peer))
              < 0) std_err();

            close(sd);

            for(i = HBMINUTES; i; i--) {
                printf(" %3d\r", i);
                sleep(ONESEC);
            }
        }
        return(0);

    } else if(hbmethod == 1) {
        printf("\n"
            "The game port %d will be temporary bound each 5 minutes, look the details:\n",
            heartbeat_port);

        peerl.sin_addr.s_addr = INADDR_ANY;
        peerl.sin_port        = htons(heartbeat_port);
        peerl.sin_family      = AF_INET;
        psz                   = sizeof(peer);

        for(;;) {
            sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
            if(sd < 0) std_err();

            fputs("Binding: ", stdout);
            if(setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))
              < 0) std_err();
            if(bind(sd, (struct sockaddr *)&peerl, sizeof(peer))
              < 0) std_err();

            len = snprintf(
                buff,
                BUFFSZ,
                GSHB1,
                heartbeat_port,
                gamestr);
            CHK_BUFFLEN;

            if(sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(peer))
              < 0) std_err();
            if(!quiet) fputc('.', stdout);

            for(;;) {
                if(timeout(sd) < 0) {
                    fputs("\n"
                        "Error: socket timeout, probably your firewall blocks the UDP packets to or from the master server\n"
                        "\n", stderr);
                    exit(1);
                }

                err = recvfrom(sd, buff, BUFFSZ, 0, (struct sockaddr *)&peer, &psz);
                if(err <= 0) continue;

                    // here we check if the source IP is the same or is in the same B
                    // subnet of the master server
                if((peer.sin_addr.s_addr & 0xffff) == (msip & 0xffff)) break;
                fputc('x', stdout);
            }
            if(!quiet) fputc('.', stdout);
            buff[err] = 0x00;

            sec = strstr(buff, "\\secure\\");
            if(sec) {
                sec += 8;
                key = gsseckey(sec, gamekey, enctype);
                printf("\n- Secure: %s  Key:    %s\n", sec, key);
                len = snprintf(
                    buff,
                    BUFFSZ,
                    GSHB2a,
                    key);
            } else {
                len = snprintf(
                    buff,
                    BUFFSZ,
                    GSHB2b,
                    gamestr);
            }
            CHK_BUFFLEN;

            if(sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(peer))
              < 0) std_err();
            if(!quiet) fputc('.', stdout);

            close(sd);
            fputs(" free\n", stdout);

            for(i = HBMINUTES; i; i--) {
                printf(" %3d\r", i);
                sleep(ONESEC);
            }

                /* needed to avoid that the sockaddr_in structure */
                /* is overwritten with a different host and port */
            peer.sin_addr.s_addr = msip;
            peer.sin_port        = htons(msport);
            peer.sin_family      = AF_INET;
        }
        return(0);
    }

    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    if(connect(sd, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    for(len = 0; len < BUFFSZ;) {
        err = recv(sd, buff + len, BUFFSZ - len, 0);
        if(err < 0) std_err();
        if(!err) break;
        len += err;
        if(!buff[len - 1]) break;
    }
    buff[len] = 0x00;

    sec = strstr(buff, "\\secure\\");
    if(!sec) {
        fprintf(stderr, "\n"
            "Alert: This master server doesn't seem to support the \"secure\" Gamespy protocol\n"
            "       This is the received reply: %s\n"
            "       I try to send the query with an empty \"validate\" value\n"
            "\n", buff);
        sec = key = "";
    } else {
        sec += 8;
        key = gsseckey(sec, gamekey, enctype);
    }

    if(!quiet) {
        fprintf(fdout,
            "Secure:      %s\n"
            "Validate:    %s\n",
            sec, key);
    }

    len = snprintf(
        buff,
        BUFFSZ,
        PCK,
        gamestr,
        enctype,
        key,
        gamestr,
        where,
        filter);
    CHK_BUFFLEN;

    if(send(sd, buff, len, 0)
      < 0) std_err();

    switch(outtype) {
        case 0: break;
        case 1:
        case 3: {
            snprintf(buff, BUFFSZ, "%s.gsl", gamestr);
            if(!quiet) printf("\nOutput file: %s\n", buff);
            fdout = fopen(buff, "wb");
            if(!fdout) std_err();
            } break;
        case 2:
        case 4: {
            if(!quiet) printf("\nOutput file: %s\n", FILEOUT);
            fdout = fopen(FILEOUT, "wb");
            if(!fdout) std_err();
            } break;
        case 5:
        case 6: {
            fdout = stdout;
            }
    }

    if(!quiet) fputs("Receiving:   ", stdout);
    len = 0;
    while((err = recv(sd, buff + len, dynsz - len, 0)) > 0) {
        if(!quiet) fputc('.', stdout);
        len += err;
        if(len >= dynsz) {
            dynsz += BUFFSZ;
            buff = realloc(buff, dynsz);
            if(!buff) std_err();
        }
    }
    close(sd);

    if(!quiet) printf(" %u bytes\n", len);

    if(outtype == 6) {
        fputs("\n\n", stdout);
        show_dump(buff, len, fdout);
        fputc('\n', stdout);
        free(buff);
        return(0);
    }

    ipport = (struct ipport_struct *)buff;
    if(enctype == 2) {
        ipport = (struct ipport_struct *)enctype2_decoder(
            gamekey,
            buff,
            &len);
    }

    if(!quiet) fputs("\n-----------------------\n", stdout);
    servers = 0;
    while(len >= 6) {
        if(!strncmp((char *)ipport, "\\final\\", 7)) {
            itsok = 1;
            break;
        }

        switch(outtype) {
            case 0: {
                fprintf(fdout,
                    "%15s   %hu\n",
                    inet_ntoa(*(struct in_addr *)&ipport->ip),
                    ntohs(ipport->port));
                } break;
            case 5:
            case 1:
            case 2: {
                fprintf(fdout,
                    "%s:%hu\n",
                    inet_ntoa(*(struct in_addr *)&ipport->ip),
                    ntohs(ipport->port));
                } break;
            case 3:
            case 4: {
                fwrite((void *)ipport, 6, 1, fdout);
                } break;
        }

            /* external program execution */
        if(execstring) {
            execptr = tmpexec + execlen;
            if(execstring_ip && !execstring_port) {
                execptr += sprintf(execptr, "%s", inet_ntoa(*(struct in_addr *)&ipport->ip));
                strcpy(execptr, execstring_ip + 3);

            } else if(execstring_port && !execstring_ip) {
                execptr += sprintf(execptr, "%d", ntohs(ipport->port));
                strcpy(execptr, execstring_port + 5);

            } else if(execstring_ip < execstring_port) {
                execptr += sprintf(execptr, "%s", inet_ntoa(*(struct in_addr *)&ipport->ip));
                execptr += sprintf(execptr, "%s", execstring_ip + 3);
                execptr += sprintf(execptr, "%d", ntohs(ipport->port));
                strcpy(execptr, execstring_port + 5);

            } else if(execstring_port < execstring_ip) {
                execptr += sprintf(execptr, "%d", ntohs(ipport->port));
                execptr += sprintf(execptr, "%s", execstring_port + 5);
                execptr += sprintf(execptr, "%s", inet_ntoa(*(struct in_addr *)&ipport->ip));
                strcpy(execptr, execstring_ip + 3);
            }

            printf("   Execute program: \"%s\"\n", tmpexec);
            system(tmpexec);
        }

        servers++;
//        ipport = (struct ipport_struct *)((char *)ipport + 6); // old method
        ipport++;
        len -= 6;
    }

    free(buff);
    if(tmpexec) free(tmpexec);
    if(outtype) fclose(fdout);

    if(!quiet) {
        if(!itsok) {
            fputs("\n\n"
                "Alert: your query seems to have not been accepted, there are at least 2\n"
                "       possible reasons: the first is that the game you have choosen has been\n"
                "       removed by Gamespy (like what happened to Judge Dredd because its\n"
                "       distributor didn't want to show the game servers of this game),\n"
                "       otherwise there is a bug or a compatibility bug in Gslist.\n"
                "       If you have doubt try the -o 6 option to see the hex dump of the\n"
                "       received server's data\n"
                "\n", stderr);
        } else {
            printf("\n\nOnline there are %u servers\n\n", servers);
        }
    }

    return(0);
}



void gs_info_udp(u_int ip, u_short port, int mode) {
    struct  sockaddr_in peer;
    int     sd,
            len,
            nt,
            chr,
            querylen;
    u_char  buff[2048],
            *query,
            newgsinfo[] =
                "\xfe\xfd\x00" "\x00\x00\x00\x00" "\xff\x00\x00";
                //   FF = on, 00 = off             |   |   |
                //                                 |   |   teams
                //                                 |   players
                //                                 informations

    // was ok also to directly use multi_query but I prefer this longer solution
    peer.sin_addr.s_addr = ip;
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    fprintf(fdout, "\nTarget   %s : %hu\n",
        inet_ntoa(peer.sin_addr), port);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    if(mode) {
        *(int *)(newgsinfo + 3) = ~time(NULL);
        query = newgsinfo;
        querylen = sizeof(newgsinfo) - 1;
        chr = 0x00;
        nt = 0;
    } else {
        query = OLDGSINFO;
        querylen = sizeof(OLDGSINFO) - 1;
        chr = '\\';
        nt = 1;
    }

    if(sendto(sd, query, querylen, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    if(timeout(sd) < 0) {
        fputs("\n"
            "Error: Socket timeout, no reply received\n"
            "\n", stderr);
        exit(1);
    }
    len = recvfrom(sd, buff, sizeof(buff) - 1, 0, NULL, NULL);
    if(len < 0) std_err();
    buff[len] = 0x00;

    if(!nt) generic_query(nt, chr, buff, len, 5, 0);
      else generic_query(nt, chr, buff, len, 0, 0);

    close(sd);
}



int check_updates(void) {
    FILE    *fdsvc,
            *fd;
    int     found = 0,
            tot   = 0;
    u_char  gamename[NAMESZ + 1],
            buff[TOTSZ + 1],
            *name,
            *p;

    fd = gslfopen(CFG, "rb");
    if(!fd) std_err();

    fdsvc = gslfopen(KNOWNCFG, "rb");
    if(!fdsvc) std_err();

    while(fgets(gamename, NAMESZ, fdsvc)) {
        if(!*gamename || (*gamename == '-')) continue;
        p = strchr(gamename, ' ');
        if(p) *p = 0x00;

        rewind(fd);
        while(fgets(buff, TOTSZ, fd)) {
            name = buff + CNAMEOFF;
            p = strchr(name, ' ');
            if(p) *p = 0x00;
            if(!strcmp(gamename, name)) {
                found++;
                break;
            }
        }
        tot++;
    }

    fclose(fd);
    fclose(fdsvc);
    return(tot - found);
}



void gsfullup(void) {
    u_char  getfile[GETSZ + 1],
            gamename[NAMESZ + 1],
            choice,
            *buff,
            *s,
            *p;
    int     err,
            i,
            updates,
            firsttime = 0;
    FILE    *fdsvc,
            *fdfull;


        /* GETNAMES */

    fputs("\n1) Start download of gamenames ("HOST":80):\n", stdout);

    fdfull = gslfopen(CFG, "rb");
    if(fdfull) {
        fputs("\nDo you want to overwrite "KNOWNCFG", "FULLCFG" and "CFG"? (y/n): ", stdout);
        fflush(stdin);
        choice = fgetc(stdin);
        if((choice != 'y') && (choice != 'Y')) {
            fputs("\n"
                "Your Gslist configuration files will not be updated\n"
                "\n", stdout);
            exit(1);
        }
        fclose(fdfull);
    } else {
        firsttime = 1;
    }
    fdsvc = gslfopen(KNOWNCFG, "wb");
    if(!fdsvc) {
        fputs("\nCreate file "KNOWNCFG, stdout);
        std_err();
    }

    fputs("\n          Connection to "HOST"\r", stdout);
    http2file(GETNAME, sizeof(GETNAME) - 1, fdsvc);
    fclose(fdsvc);
    fputs("\nGamenames downloaded successfully\n", stdout);

        /* CHECK FOR UPDATES */

    if(!firsttime) {
        fputs("\nCheck for updates...\n", stdout);
        updates = check_updates();
        if(!updates) {
            fputs("\nNo updates or new games available\n\n", stdout);
            return;
        }
        printf("\n%d updates available\n", updates);
    }

        /* GET INFO */

    fputs("\n2) Start download of "FULLCFG" ("HOST":80):\n", stdout);

    fputs("\nOpen "KNOWNCFG"\n", stdout);
    fdsvc = gslfopen(KNOWNCFG, "rb");
    if(!fdsvc) std_err();

    fputs("\nCreate "FULLCFG"\n", stdout);
    fdfull = gslfopen(FULLCFG, "wb");
    if(!fdfull) std_err();

    buff = (u_char *)memcpy(getfile, GET1, sizeof(GET1) - 1)
           + (sizeof(GET1) - 1);
    while(!feof(fdsvc)) {
        p = buff;
        i = 0;
        for(;;) {
            s = fgets(gamename, NAMESZ, fdsvc);
            if(!s) {
                p--;
                *p = 0x00;
                break;
            }
            s = strchr(gamename, ' ');
            if(s) {
                *s = 0x00;
                err = s - gamename;
            } else {
                err = strlen(gamename);
                // s = gamename + err;
            }

            if(*gamename == '-') continue;  // -gamename means the game no longer exists
            p = (u_char *)memcpy(p, gamename, err) + err;
            if(i == MAXNAMEREQ) {
                *p = 0x00;
                break;
            }
            *p++ = '\\';
            i++;
        }

        err = (p - buff) + sizeof(GET1) - 1;
        memcpy(getfile + err, GET2, sizeof(GET2));
        err += sizeof(GET2) - 1;
        fputs("\n          Parts of "FULLCFG": Connection to "HOST":\r", stdout);
        http2file(getfile, err, fdfull);
    }
    fclose(fdsvc);
    fclose(fdfull);

    fputs("\nUpdate has been completed successfully!!!\n\n", stdout);

    build_cfg();

    if(firsttime) {
        fputs("\n"
            "The configuration file of Gslist ("CFG") has been created\n"
            "It is a text file so you can open it with any text editor\n"
            "The file "KNOWNCFG" and "FULLCFG" instead now are no longer needed, so you can also delete them but I think the data in them is enough interesting so check them if you like\n"
            "\n", stdout);
    }
}



void http2file(u_char *getstr, int len, FILE *fd) {
    int     sock,
            err,            // temporary length
            length   = 0,   // total length
            length2  = 0,   // used only for the HTTP header
            realfile = 0;
    u_int   fsize    = 0,
            perc     = 0;
    u_char  filebuff[GETSZ + 1],
            *p,
            *contlen,
            *page;
    struct  sockaddr_in peer;


    peer.sin_addr.s_addr = resolv(HOST);
    peer.sin_port        = htons(PORT);
    peer.sin_family      = AF_INET;

    sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sock < 0) std_err();

    if(connect(sock, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    if(send(sock, getstr, len, 0)
      < 0) std_err();

    while((err = recv(sock, filebuff + length2, GETSZ - length2, 0)) > 0) {
        length += err;

        if(realfile) {
            if(fwrite(filebuff, err, 1, fd) != 1) write_err();
            if(fsize) printf("%03u%%\r", length / perc);
                else fputc('.', stdout);
        } else {
            filebuff[length] = 0x00;
            p = strstr(filebuff, ENDHEAD);
            if(p) {
                p += (sizeof(ENDHEAD) - 1);

                page = filebuff + 9;
                if(!memcmp(page, H416, sizeof(H416) - 1)) {
                    close(sock);
                    return;
                }
                if(memcmp(page, H200, sizeof(H200) - 1) &&
                   memcmp(page, H206, sizeof(H206) - 1)) {
                    fputs("\nError: File not found on the server\n", stderr);
                    close(sock);
                    return;
                }
                contlen = stristr(filebuff, CLEN);
                if(contlen) {
                    fsize = atol(contlen + sizeof(CLEN) - 1);
                    perc = fsize / 100;
                }
                length2 = 0;
                length += filebuff - p;
                if(fwrite(p, length, 1, fd) != 1) write_err();
                realfile = 1;
            } else {
                length2 = length;
            }
        }
    }

    close(sock);
}



void write_err(void) {
    fputs("\n"
        "Error: I/O error. Probably your disk is full or the file is write protected\n"
        "\n", stderr);
    exit(1);
}



void build_cfg(void) {
    FILE    *fd,
            *fdcfg;
    u_char  buff[TOTSZ + 1],
            tot[TOTSZ + 1],
            *p,
            *tmp;
    int     len,
            i = 1;

    fd = gslfopen(FULLCFG, "rb");
    if(!fd) std_err();

    fputs("Build file "CFG"\n", stdout);
    fdcfg = gslfopen(CFG, "wb");
    if(!fdcfg) std_err();

    *tot = 0x00;
    while(fgets(buff, TOTSZ, fd)) {
        if(*buff == '[') {      // gamename = CNAME
            fputs(tot, fdcfg);
            fputc('\n', fdcfg);

            tmp = buff + 1;
            tmp[CNAMELEN - 1] = 0x00;
            p = strchr(tmp, ']');
            if(p) *p = 0x00;
            len = sprintf(tot, "%4d)  %s", i, tmp);
            memset(tot + len, ' ', CFNAMEOFF - len);
            tot[CFNAMEOFF] = 0x00;
            tot[CKEYOFF] = 0x00;
            i++;

        } else if(!strnicmp(buff, "fullname=", 9)) {    // CFNAME
            tmp = buff + 9;
            tmp[CFNAMELEN - 1] = 0x00;
            for(p = tmp; *p > '\r'; p++);
            len = p - tmp;
            p = (u_char *)memcpy(tot + CFNAMEOFF, tmp, len) + len;
            memset(p, ' ', CSPACES - len);

        } else if(!strnicmp(buff, "handoff=", 8)) {
            p = tot + CKEYOFF;
            *p++ = buff[10];
            *p++ = buff[12];
            *p++ = buff[14];
            *p++ = buff[16];
            *p++ = buff[18];
            *p++ = buff[20];
            *p = 0x00;
        }
    }
    fputs(tot, fdcfg);
    fputc('\n', fdcfg);

    fclose(fd);
    fclose(fdcfg);
}



void get_key(int gamenum, u_char *gamestr, u_char *gamekey) {
    FILE    *fd;
    u_char  buff[TOTSZ + 1],
            *s;
    int     i;

    *gamekey = 0x00;

    fd = gslfopen(CFG, "rb");
    if(!fd) {
        fputs("\n"
            "Open file "CFG"\n"
            "Error: you must launch \"gslist -u\" the first time!!!",
            stderr);
        std_err();
    }
    if(gamenum) {
        for(i = 0; i <= gamenum; i++) {
            fgets(buff, TOTSZ, fd);
        }
        buff[CFNAMEOFF] = 0x00;
        sscanf(buff + CNAMEOFF, "%s", gamestr);
        memcpy(gamekey, buff + CKEYOFF, 6);
        gamekey[6] = 0x00;
    } else {
        while(fgets(buff, TOTSZ, fd)) {
            s = strchr(buff + CNAMEOFF, ' ');
            if(s) *s = 0x00;
            if(!strcmp(gamestr, buff + CNAMEOFF)) {
                memcpy(gamekey, buff + CKEYOFF, 6);
                gamekey[6] = 0x00;
                break;
            }
        }
    }
    fclose(fd);

    if((*gamekey == ' ') || (*gamekey == '\n') || (*gamekey == '\t')) {
        *gamekey = 0x00;
    }
}



void show_list(u_char *name) {
    FILE    *fd;
    u_char  buff[TOTSZ + 1];

    fd = gslfopen(CFG, "rb");
    if(!fd) {
        fputs("\n"
            "Open file "CFG"\n"
            "Error: you must launch \"gslist -u\" the first time!\n"
            "       Is also possible that you are launching Gslist from a directory while\n"
            "       the configuration files are in another one",
            stderr);
        std_err();
    }

    if(!quiet) {
        fputs(" NUM)  GAMENAME        FULL GAME NAME                                    KEY\n"
              "-------------------------------------------------------------------------------",
              fdout);
    }

    if(name) {
        fputc('\n', stdout);
        while(fgets(buff, TOTSZ, fd)) {
            if(stristr(buff, name)) fputs(buff, fdout);
        }
    } else {
        while(fgets(buff, TOTSZ, fd)) {
            fputs(buff, fdout);
        }
    }
    fclose(fd);
}



void show_countries(void) {
    const static char   *countries[][2] = {
        { "[Codes]", "" },
        { "US", "United States" },
        { "ZZ", "Other" },
        { "AF", "Afghanistan" },
        { "AL", "Albania" },
        { "DZ", "Algeria" },
        { "AS", "American Samoa" },
        { "AD", "Andorra" },
        { "AO", "Angola" },
        { "AI", "Anguilla" },
        { "AQ", "Antarctica" },
        { "AG", "Antigua and Barbuda" },
        { "AR", "Argentina" },
        { "AM", "Armenia" },
        { "AW", "Aruba" },
        { "AU", "Australia" },
        { "AT", "Austria" },
        { "AZ", "Azerbaijan" },
        { "BS", "Bahamas" },
        { "BH", "Bahrain" },
        { "BD", "Bangladesh" },
        { "BB", "Barbados" },
        { "BY", "Belarus" },
        { "BE", "Belgium" },
        { "BZ", "Belize" },
        { "BJ", "Benin" },
        { "BM", "Bermuda" },
        { "BT", "Bhutan" },
        { "BO", "Bolivia" },
        { "BA", "Bosnia and Herzegovina" },
        { "BW", "Botswana" },
        { "BV", "Bouvet Island" },
        { "BR", "Brazil" },
        { "IO", "British Indian Ocean Territory" },
        { "BN", "Brunei Darussalam" },
        { "BG", "Bulgaria" },
        { "BF", "Burkina Faso" },
        { "BI", "Burundi" },
        { "KH", "Cambodia" },
        { "CM", "Cameroon" },
        { "CA", "Canada" },
        { "CV", "Cape Verde" },
        { "CT", "Catalonia" },
        { "KY", "Cayman Islands" },
        { "CF", "Central African Republic" },
        { "TD", "Chad" },
        { "CL", "Chile" },
        { "CN", "China" },
        { "CX", "Christmas Island" },
        { "CC", "Cocos (Keeling Islands)" },
        { "CO", "Colombia" },
        { "KM", "Comoros" },
        { "CG", "Congo" },
        { "CK", "Cook Islands" },
        { "CR", "Costa Rica" },
        { "CI", "Cote D'Ivoire (Ivory Coast)" },
        { "HR", "Croatia (Hrvatska)" },
        { "CU", "Cuba" },
        { "CY", "Cyprus" },
        { "CZ", "Czech Republic" },
        { "DK", "Denmark" },
        { "DJ", "Djibouti" },
        { "DM", "Dominica" },
        { "DO", "Dominican Republic" },
        { "TP", "East Timor" },
        { "EC", "Ecuador" },
        { "EG", "Egypt" },
        { "SV", "El Salvador" },
        { "GQ", "Equatorial Guinea" },
        { "ER", "Eritrea" },
        { "EE", "Estonia" },
        { "ET", "Ethiopia" },
        { "FK", "Falkland Islands (Malvinas)" },
        { "FO", "Faroe Islands" },
        { "FJ", "Fiji" },
        { "FI", "Finland" },
        { "FR", "France" },
        { "FX", "France Metropolitan" },
        { "GF", "French Guiana" },
        { "PF", "French Polynesia" },
        { "TF", "French Southern Territories" },
        { "GA", "Gabon" },
        { "GM", "Gambia" },
        { "GE", "Georgia" },
        { "DE", "Germany" },
        { "GH", "Ghana" },
        { "GI", "Gibraltar" },
        { "GR", "Greece" },
        { "GL", "Greenland" },
        { "GD", "Grenada" },
        { "GP", "Guadeloupe" },
        { "GU", "Guam" },
        { "GT", "Guatemala" },
        { "GN", "Guinea" },
        { "GW", "Guinea-Bissau" },
        { "GY", "Guyana" },
        { "HT", "Haiti" },
        { "HM", "Heard and McDonald Islands" },
        { "HN", "Honduras" },
        { "HK", "Hong Kong" },
        { "HU", "Hungary" },
        { "IS", "Iceland" },
        { "IN", "India" },
        { "ID", "Indonesia" },
        { "IR", "Iran" },
        { "IQ", "Iraq" },
        { "IE", "Ireland" },
        { "IL", "Israel" },
        { "IT", "Italy" },
        { "JM", "Jamaica" },
        { "JP", "Japan" },
        { "JO", "Jordan" },
        { "KZ", "Kazakhstan" },
        { "KE", "Kenya" },
        { "KI", "Kiribati" },
        { "KP", "North Korea" },
        { "KR", "South Korea" },
        { "KW", "Kuwait" },
        { "KG", "Kyrgyzstan" },
        { "LA", "Laos" },
        { "LV", "Latvia" },
        { "LB", "Lebanon" },
        { "LS", "Lesotho" },
        { "LR", "Liberia" },
        { "LY", "Libya" },
        { "LI", "Liechtenstein" },
        { "LT", "Lithuania" },
        { "LU", "Luxembourg" },
        { "MO", "Macau" },
        { "MK", "Macedonia" },
        { "MG", "Madagascar" },
        { "MW", "Malawi" },
        { "MY", "Malaysia" },
        { "MV", "Maldives" },
        { "ML", "Mali" },
        { "MT", "Malta" },
        { "MH", "Marshall Islands" },
        { "MQ", "Martinique" },
        { "MR", "Mauritania" },
        { "MU", "Mauritius" },
        { "YT", "Mayotte" },
        { "MX", "Mexico" },
        { "FM", "Micronesia" },
        { "MD", "Moldova" },
        { "MC", "Monaco" },
        { "MN", "Mongolia" },
        { "MS", "Montserrat" },
        { "MA", "Morocco" },
        { "MZ", "Mozambique" },
        { "MM", "Myanmar (Burma)" },
        { "NA", "Namibia" },
        { "NR", "Nauru" },
        { "NP", "Nepal" },
        { "NL", "Netherlands" },
        { "AN", "Netherlands Antilles" },
        { "NC", "New Caledonia" },
        { "NZ", "New Zealand" },
        { "NI", "Nicaragua" },
        { "NE", "Niger" },
        { "NG", "Nigeria" },
        { "NU", "Niue" },
        { "NF", "Norfolk Island" },
        { "MP", "Northern Mariana Islands" },
        { "NO", "Norway" },
        { "OM", "Oman" },
        { "PK", "Pakistan" },
        { "PW", "Palau" },
        { "PA", "Panama" },
        { "PG", "Papua New Guinea" },
        { "PY", "Paraguay" },
        { "PE", "Peru" },
        { "PH", "Philippines" },
        { "PN", "Pitcairn" },
        { "PL", "Poland" },
        { "PT", "Portugal" },
        { "PR", "Puerto Rico" },
        { "QA", "Qatar" },
        { "RE", "Reunion" },
        { "RO", "Romania" },
        { "RU", "Russian Federation" },
        { "RW", "Rwanda" },
        { "KN", "Saint Kitts and Nevis" },
        { "LC", "Saint Lucia" },
        { "VC", "Saint Vincent and The Grenadines" },
        { "WS", "Samoa" },
        { "SM", "San Marino" },
        { "ST", "Sao Tome and Principe" },
        { "SA", "Saudi Arabia" },
        { "SN", "Senegal" },
        { "SC", "Seychelles" },
        { "SL", "Sierra Leone" },
        { "SG", "Singapore" },
        { "SK", "Slovak Republic" },
        { "SI", "Slovenia" },
        { "SB", "Solomon Islands" },
        { "SO", "Somalia" },
        { "ZA", "South Africa" },
        { "GS", "S. Georgia and S. Sandwich Isls." },
        { "ES", "Spain" },
        { "LK", "Sri Lanka" },
        { "SH", "St. Helena" },
        { "PM", "St. Pierre and Miquelon" },
        { "SD", "Sudan" },
        { "SR", "Suriname" },
        { "SJ", "Svalbard and Jan Mayen Islands" },
        { "SZ", "Swaziland" },
        { "SE", "Sweden" },
        { "CH", "Switzerland" },
        { "SY", "Syria" },
        { "TW", "Taiwan" },
        { "TJ", "Tajikistan" },
        { "TZ", "Tanzania" },
        { "TH", "Thailand" },
        { "TG", "Togo" },
        { "TK", "Tokelau" },
        { "TO", "Tonga" },
        { "TT", "Trinidad and Tobago" },
        { "TN", "Tunisia" },
        { "TR", "Turkey" },
        { "TM", "Turkmenistan" },
        { "TC", "Turks and Caicos Islands" },
        { "TV", "Tuvalu" },
        { "UG", "Uganda" },
        { "UA", "Ukraine" },
        { "AE", "United Arab Emirates" },
        { "UK", "United Kingdom" },
        { "UM", "US Minor Outlying Islands" },
        { "UY", "Uruguay" },
        { "UZ", "Uzbekistan" },
        { "VU", "Vanuatu" },
        { "VA", "Vatican City" },
        { "VE", "Venezuela" },
        { "VN", "VietNam" },
        { "VG", "Virgin Islands (British)" },
        { "VI", "Virgin Islands (US)" },
        { "WF", "Wallis and Futuna Islands" },
        { "EH", "Western Sahara" },
        { "YE", "Yemen" },
        { "YU", "Yugoslavia" },
        { "ZR", "Zaire" },
        { "ZM", "Zambia" },
        { "ZW", "Zimbabwe" },
        { "", "" },
        { "[USStates]", "" },
        { "01", "Alabama " },
        { "02", "Alaska " },
        { "04", "Arizona " },
        { "05", "Arkansas " },
        { "06", "California " },
        { "08", "Colorado " },
        { "09", "Connecticut " },
        { "10", "Delaware " },
        { "11", "Dist. of Columbia " },
        { "12", "Florida " },
        { "13", "Georgia " },
        { "15", "Hawaii " },
        { "16", "Idaho " },
        { "17", "Illinois " },
        { "18", "Indiana " },
        { "19", "Iowa " },
        { "20", "Kansas " },
        { "21", "Kentucky " },
        { "22", "Louisiana " },
        { "23", "Maine " },
        { "24", "Maryland " },
        { "25", "Massachusetts " },
        { "26", "Michigan " },
        { "27", "Minnesota " },
        { "28", "Mississippi " },
        { "29", "Missouri " },
        { "30", "Montana " },
        { "31", "Nebraska " },
        { "32", "Nevada " },
        { "33", "New Hampshire " },
        { "34", "New Jersey " },
        { "35", "New Mexico " },
        { "36", "New York " },
        { "37", "North Carolina " },
        { "38", "North Dakota " },
        { "39", "Ohio " },
        { "40", "Oklahoma " },
        { "41", "Oregon " },
        { "42", "Pennsylvania " },
        { "44", "Rhode Island " },
        { "45", "South Carolina " },
        { "46", "South Dakota " },
        { "47", "Tennessee " },
        { "48", "Texas " },
        { "49", "Utah " },
        { "50", "Vermont " },
        { "51", "Virginia " },
        { "53", "Washington " },
        { "54", "West Virginia " },
        { "55", "Wisconsin " },
        { "56", "Wyoming " },
        { "60", "American Samoa " },
        { "64", "Micronesia " },
        { "66", "Guam " },
        { "68", "Marshall Islands " },
        { "69", "No. Mariana Islands  " },
        { "70", "Palau " },
        { "72", "Puerto Rico " },
        { "74", "U.S. Islands" },
        { "78", "Virgin Islands" },
        { NULL, NULL }
    };
    int i;

    for(i = 0; countries[i][0]; i++) {
        printf("%s   %s\n", countries[i][0], countries[i][1]);
    }
}



void show_help(void) {
    show_gslist_title();

    printf("\n"
        "Usage: gslist <options>\n"
        "\n"
        "Options:\n"
        "-e             some usage examples\n"
        "-N GAMENAME    retrieves the servers list of the game called exactly GAMENAME\n"
        "               (use -l or -s to know gamename). Example: gslist -N ut2\n"
        "-n NUM         obsolete option to retrieve the servers list using their places\n"
        "               (use -l or -s to know the left number). Example: gslist -n 166\n"
        "-l             Shows the complete list of supported games and their details\n"
        "-s PATTERN     searchs the game name, parts of the name or any other things\n"
        "               into the list of supported games and shows some details.\n"
        "               It is case insensitve, so -s UNREAL and -s unReaL are the same\n"
        "-u             updates the current list of supported games in "CFG"\n"
        "               directly contacting the Gamespy webserver (use it for new games)\n"
        "-i HOST PORT   requests informations to a remote game that supports the old\n"
        "               Gamespy query protocol ("OLDGSINFO")\n"
        "-I HOST PORT   as above but using the new Gamespy query protocol (FE FD ...)\n"
        MULTIQUERY
        "-f FILTERS     specify a filter to apply to the servers list before returning.\n"
        "               Valid operators are:\n"
        "                 <>, !=, >=, !<, <=, !>, =, <, >, (, ), +, -, *, /, %%,\n"
        "                 AND, NOT, OR, LIKE, NOT LIKE, IS NULL, IS NOT NULL\n"
        "               Valid items are:\n"
        "                 hostaddr, hostport, gamever, country (check the -c option)\n"
        "                 hostname, mapname, gametype, gamemode, numplayers, maxplayers\n"
        "               Wildcard character: %%     String delimiter:   '\n"
        "               Examples: -f \"((hostname LIKE '%%Bob%%') AND (country='US'))\"\n"
        "                         -f \"(numplayers > 0)\"\n"
        "-c             shows the list of country codes to use with the filter \"country\"\n"
        "-r \"prog...\"   lets you to execute a specific program for each IP found (as the\n"
        "               ping command for example).\n"
        "               There are 2 available parameters and are #IP and #PORT that will\n"
        "               be substituited with the IP and port of the current online game\n"
        "               server found in the list. Example: -r \"echo #IP and #PORT yeah\"\n"
        "-b PORT        heartbeats sender, lets your IP address to be included in the\n"
        "               online servers list, the PORT is the query port of your server.\n"
        "               PORT will be bound for some milliseconds to be able to answer to\n"
        "               the master server queries or possible validation requests, so\n"
        "               during this small time your server will be no accessible\n"
        "               Must be used with one of the -N, -n or -y options\n"
        "               Then note this option is a bit experimental, use -B if possible\n"
        "-B PORT        as above but the port is not bound, is needed that your server\n"
        "               supports the old Gamespy query protocol ("OLDGSINFO") otherwise\n"
        "               your server will not be added in the servers list\n"
        "-x S[:P]       lets you to specify a different master server (S) and port (P is\n"
        "               optional) instead of "MS":%d\n"
        "-o OUT         lets you to specify a different output for the servers list\n"
        "               (default output is screen). You can use:\n"
        "               1 = text output to a file for each game (ex: serioussam.gsl).\n"
        "                   string: 1.2.3.4:1234 (plus a final line-feed)\n"
        "               2 = text output as above but to only one file ("FILEOUT")\n"
        "               3 = binary output to a file for each game: 4 bytes IP, 2 port\n"
        "                   example: (hex) 7F0000011E62 = 127.0.0.1 7778\n"
        "               4 = binary output as above but to only one file ("FILEOUT")\n"
        "               5 = exactly like 1 but to stdout\n"
        "               6 = hexadecimal visualization of the raw servers list as is\n"
        "               FILENAME = if OUT is a filename all the screen output will be\n"
        "                   dumped into the file FILENAME\n"
        "-q             quiet output\n"
        "-y NAME KEY    Uses a custom gamename and game key: useful if you use Gslist\n"
        "               for one game only or if the game or applications not listed.\n"
        "               Some examples are:   -y gamespy2 d4kZca   -y gslive Xn221z\n"
        "               Check here:   http://aluigi.altervista.org/papers/gshkeys.txt\n"
        "-t NUM         enctype selection, only 0 (default) and 2 are supported\n"
        "-v             shows the version of Gslist you are using\n"
        "\n", MSPORT);
}



void show_examples(void) {
    fputs("\n"
        "Some examples of the Gslist's usage:\n"
        "\n"
        "-N ut2              retrieves the list of UT2003'servers\n"
        "-n 168              retrieves the list of UT2003'servers (168 is the\n"
        "                    current identifier of UT2003, but it can change when you\n"
        "                    update the games'list with the -u option)\n"
        "-y ut2 Lw7x0R       retrieves the list of UT2003'servers\n"
        "-N ut2 -f \"country=IT\" to retrieve all the italian servers\n"
        "-q -N ut2           retrieves the list of UT2003'servers but shows less info\n"
        "-o 1 -N ut2         the servers'list will be dumped in the file ut2.gls\n"
        "-o 2 -N ut2         the servers'list will be dumped in "FILEOUT"\n"
        "-o 3 -N ut2         the servers'list will be saved in binary form in ut2.gls\n"
        "-o 4 -N ut2         same as above but in the file "FILEOUT"\n"
        "-o x.txt -N ut2     saves the output in the file x.txt (just as you see it on\n"
        "                    screen but without errors and headers), like > x.txt\n"
        "-o 6                full hexadecimal visualization of the raw servers list\n"
        "-l                  shows the list of all the available games\n"
        "-s serious          shows all the games having a name containing \"serious\"\n"
        "-i 1.2.3.4 1234     uses the Gamespy query protocol to retrieve remote info\n"
        "                    (not all the games support this protocol)\n"
        "-I 1.2.3.4 1234     as above but uses the new Gamespy protocol\n"
        "-d 1 1.2.3.4 1234   as above but uses a Quake 3 query\n"
        "-u                  this is the great function to update the list of supported\n"
        "                    games. USE IT FREQUENTLY (moreover when a new game is out)\n"
        "-x localhost -n 2   retrieves the servers'list of the game identified by the\n"
        "                    number 2 using the masterserver at localhost:28900\n"
        "                    If you use -x localhost:1234 will be contacted the master\n"
        "                    server port 1234\n"
        "-o xxx.txt -l       the games'list will be dumped in xxx.txt\n"
        "-r \"ping -n 1 #IP\"  will ping each IP found\n"
        "-r \"gsinfo #IP #PORT\" will launch the program gsinfo\n"
        "-b 7778 -N ut2      adds your IP to the game list of Unreal Tournament 2003\n"
        "                    your port 7778 will be temporary bound and your server\n"
        "                    will not be reacheable for some milliseconds\n"
        "-B 7778 -N ut2      as above but no ports will be bound. Your server must\n"
        "                    support the "OLDGSINFO" query\n"
        "-c                  shows the list of country codes existent\n"
        "-N ut2 -t 2         uses the enctype 2 (encrypted server list), is just like\n"
        "                    enctype 0. You don't need to use this option\n"
        "-v                  shows the version of Gslist you are using\n"
        "\n"
        "I think these examples are enough, however contact me if you have problems\n"
        "\n", stdout);
}



void show_gslist_title(void) {
    fputs("\n"
        "Gslist "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    http://aluigi.altervista.org\n"
        "\n", stdout);
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            fprintf(stderr, "\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }

    FILE *gslfopen(const char *path, const char *mode) {
        int         len;
        char        fullpath[PATH_MAX + 1];
        const char  *home;

        home = getenv("HOME");
        if(!home) {
            fputs("\n"
                "Error: impossible to know your $HOME directory where to read/write the Gslist configuration files.\n"
                "       Try to modify the source code of the program or report the problem to me.\n"
                "\n", stderr);
            exit(1);
        }

        len = snprintf(
            fullpath,
            PATH_MAX,
            "%s/.gslist/",
            home);
        if((len < 0) || (len > PATH_MAX)) {
            errno = ECANCELED;
            return(NULL);
        }

        mkdir(fullpath, 0755);

        strncpy(fullpath + len, path, PATH_MAX - len);
        fullpath[PATH_MAX] = 0x00;

        return(fopen(fullpath, mode));
    }
#else
    FILE *gslfopen(const char *path, const char *mode) {
        return(fopen(path, mode));
    }
#endif


